/*  fgwirc - fungw based minimal irc client
    Copyright (C) 2019  Tibor 'Igor2' Palinkas

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along
    with this program; if not, write to the Free Software Foundation, Inc.,
    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

    Project page and contact: http://repo.hu/projects/fgwirc
*/

#include <genvector/gds_char.h>

#include "action.h"
#include "fgwirc.h"
#include "irc.h"
#include "input.h"

#include "conf.h"

#include "fgwirc_act.h"

static fgw_obj_t *fgwircobj;

static fgw_error_t act_quit(fgw_arg_t *res, int argc, fgw_arg_t *argv)
{
	gds_t tmp = {0};

	ACTION_HELP((
		"quit - quit from fgwirc",
		"Syntax: quit [quit_message]",
		"",
		"If quit message is specified, send that with the quit command to each irc server connected.",
		NULL));


	gds_append_str(&tmp, "quit :");
	ACT_MAY_GDS_REST(&tmp, 1);
	if (argc < 2)
		gds_append_str(&tmp, "Leaving.");
	irc_send_all_raw(tmp.array);
	gds_uninit(&tmp);
	fgwirc_exit = 1;
	return 0;
}

static fgw_error_t act_bind(fgw_arg_t *res, int argc, fgw_arg_t *argv)
{
	char *smod, *script;
	int mod = 0, keycode;

	ACTION_HELP((
		"bind - bind an action script one-liner to a hotkey",
		"Syntax: bind [modifiers] <keycode> <script>",
		"",
		"Modifiers are: ctrl, alt, meta (for the command key), pad (for numpad), pad2, pad3, func",
		"",
		"Example: bind ctrl 12 \"window redraw\"",
		"binds ^L to redraw the current window using the window action.",
		NULL));

	if (argc == 3) {
		ACT_CONVARG(1, FGW_INT, keycode = argv[1].val.nat_int);
		ACT_CONVARG(2, FGW_STR, script = argv[2].val.str);
	}
	else if (argc == 4) {
		ACT_CONVARG(1, FGW_STR, smod = argv[1].val.str);
		ACT_CONVARG(2, FGW_INT, keycode = argv[2].val.nat_int);
		ACT_CONVARG(3, FGW_STR, script = argv[3].val.str);
		mod = input_name2mod(smod);
	}
	else {
		ACT_ERROR("bind", "Wrong number of arguments");
		return FGW_ERR_ARGC;
	}

	if ((mod == IMOD_invalid) || (keycode < 0)  || (keycode >= INPUT_MAX_KEYCODE)) {
		ACT_ERROR("bind", "Invalid modifier or keycode");
		return FGW_ERR_ARG_CONV;
	}

	free(input_key_binding[mod][keycode]);
	input_key_binding[mod][keycode] = fgw_strdup(script);

	return 0;
}

static fgw_error_t act_echo(fgw_arg_t *res, int argc, fgw_arg_t *argv)
{
	win_t *curr = idx2win(win_curr_idx);
	gds_t tmp = {0};

	ACTION_HELP((
		"echo - print a line of static text in the current window (and log)",
		"Syntax: echo <text>",
		NULL));

	ACT_GDS_REST(&tmp, 1);
	win_append(curr, tmp.array);
	gds_uninit(&tmp);
	return 0;
}

static fgw_error_t act_set(fgw_arg_t *res, int argc, fgw_arg_t *argv)
{
	win_t *curr = idx2win(win_curr_idx);
	char *key, *val = NULL, *tname;
	char tmp[512], cval[128];

	ACTION_HELP((
		"set - change the value of a runtime config node",
		"Syntax: set <conf_node_name> <new_value>",
		NULL));

	ACT_CONVARG(1, FGW_STR, key = argv[1].val.str);
	ACT_MAY_CONVARG(2, FGW_STR, val = argv[2].val.str);

	if (val == NULL) {
		conf_item_t *c = conf_get(key);
		if (c != NULL) {
			switch(c->type) {
				case FGW_STR: tname = "string"; val = *(char **)c->var; break;
				case FGW_INT: tname = "int"; val = cval; sprintf(cval, "%d", *(int *)c->var); break;
				default:      tname = "*unknown*"; break;
			}
			sprintf(tmp, "Setting %s{%s} = %s", key, tname, val);
			win_append(curr, tmp);
			win_append(curr, c->desc);
		}
		else {
			sprintf(tmp, "Setting '%s' does not exist\n", key);
			win_append(curr, tmp);
		}
	}
	else if (conf_set(key, val) != 0) {
		ACT_ERROR("set", "Set failed");
		return -1;
	}

	return 0;
}

static fgw_error_t act_conf_get(fgw_arg_t *res, int argc, fgw_arg_t *argv)
{
	char *key;
	conf_item_t *c;

	ACTION_HELP((
		"conf_get - return the value of a config node (for scripting)",
		"Syntax: conf_get <conf_node_name>",
		NULL));

	ACT_CONVARG(1, FGW_STR, key = argv[1].val.str);

	c = conf_get(key);
	if (c == NULL)
		return FGW_ERR_ARG_CONV;

	switch(c->type) {
		case FGW_STR: res->type = FGW_STR; res->val.str = *(char **)c->var; break;
		case FGW_INT: res->type = FGW_INT; res->val.nat_int = *(int *)c->var; break;
		default:      return FGW_ERR_ARG_CONV;
	}

	return 0;
}

static fgw_error_t act_toggle(fgw_arg_t *res, int argc, fgw_arg_t *argv)
{
	char *key;
	conf_item_t *c;
	int changed = 0;

	ACTION_HELP((
		"toggle - toggle the value of a boolean (integer) config node",
		"Syntax: toggle <conf_node_name>",
		NULL));

	ACT_CONVARG(1, FGW_STR, key = argv[1].val.str);

	c = conf_get(key);
	switch(c->type) {
		case FGW_STR: ACT_ERROR("toggle", "Can not toggle string"); break;
		case FGW_INT: *(int *)c->var = !*(int *)c->var; changed = 1; break;
		default:      ACT_ERROR("toggle", "Can not toggle unknown type"); break;
	}

	if (changed && (c->on_chg != NULL))
		c->on_chg(c);

	return 0;
}

static fgw_error_t act_input_meta(fgw_arg_t *res, int argc, fgw_arg_t *argv)
{
	ACTION_HELP((
		"input_meta - press the meta key (a.k.a. the command key)",
		NULL));

	input_meta = 1;
	return 0;
}

static fgw_error_t act_help(fgw_arg_t *res, int argc, fgw_arg_t *argv)
{
	char *act = "help";

	ACTION_HELP((
		"help - print help about a specific action",
		"Syntax: help <action>",
		NULL));

	ACT_MAY_CONVARG(1, FGW_STR, act = argv[1].val.str);

	action_vstr(act, ACTION_HELP_STR, NULL);
	return 0;
}

static int actsort(const void *a, const void *b) { return strcmp(*(char **)a, *(char **)b); }

static fgw_error_t act_actlist(fgw_arg_t *res, int argc, fgw_arg_t *argv)
{
	htsp_entry_t *e;
	vtp0_t sorted = {0};
	gds_t buf = {0};
	win_t *curr = idx2win(win_curr_idx);
	long n;

	ACTION_HELP((
		"actlist - list all actions available",
		"Syntax: actlist",
		NULL));

	for(e = htsp_first(&fgw.func_tbl); e != NULL; e = htsp_next(&fgw.func_tbl, e))
		if (strchr(e->key, '.') == 0)
			vtp0_append(&sorted, e->key);

	qsort(sorted.array, sorted.used, sizeof(char *), actsort);

	for(n = 0; n < sorted.used; n++) {
		if (buf.used > 0)
			gds_append(&buf, ' ');
		gds_append_str(&buf, sorted.array[n]);
		if (buf.used > 500) {
			win_append(curr, buf.array);
			buf.used = 0;
		}
	}

	if (buf.used > 0)
		win_append(curr, buf.array);

	vtp0_uninit(&sorted);
	gds_uninit(&buf);

	return 0;
}

void fgwirc_act_init(void)
{
	fgwircobj = fgw_obj_reg(&fgw, "fgwirc");

	fgw_func_reg(fgwircobj, "quit", act_quit);
	fgw_func_reg(fgwircobj, "exit", act_quit);
	fgw_func_reg(fgwircobj, "bind", act_bind);
	fgw_func_reg(fgwircobj, "echo", act_echo);
	fgw_func_reg(fgwircobj, "set",  act_set);
	fgw_func_reg(fgwircobj, "toggle",  act_toggle);
	fgw_func_reg(fgwircobj, "conf_get",  act_conf_get);
	fgw_func_reg(fgwircobj, "input_meta",  act_input_meta);
	fgw_func_reg(fgwircobj, "help",  act_help);
	fgw_func_reg(fgwircobj, "actlist",  act_actlist);
}
