/*  fgwirc - fungw based minimal irc client
    Copyright (C) 2019,2021  Tibor 'Igor2' Palinkas

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along
    with this program; if not, write to the Free Software Foundation, Inc.,
    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

    Project page and contact: http://repo.hu/projects/fgwirc
*/

#include <ctype.h>
#include <genvector/gds_char.h>

#include "action.h"
#include "channel.h"
#include "fgwirc.h"
#include "irc.h"
#include "target.h"

#include "fgwirc_act.h"

static fgw_obj_t *ircobj;

static fgw_error_t act_join(fgw_arg_t *res, int argc, fgw_arg_t *argv)
{
	win_t *curr = idx2win(win_curr_idx);
	char *chan = NULL, *key = NULL;

	ACTION_HELP((
		"join - irc command: join a channel",
		"Syntax:",
		"  join <#chan>      - join #chan on server associated with current window",
		"  join <id>/<#chan> - join #chan on server id and associate current window",
		"  join              - rejoin last channel used by the window (e.g. after part)",
		NULL));

	ACT_MAY_CONVARG(1, FGW_STR, chan = argv[1].val.str);
	ACT_MAY_CONVARG(2, FGW_STR, key = argv[2].val.str);

	if (chan == NULL) {
		chan = fgw_strdup(curr->target.name);
		if ((chan == NULL) || (!irc_is_prefix_chan(*chan))) {
			ACT_ERROR("join", "No previous channel found");
			return FGW_ERR_ARG_CONV;
		}
		key = curr->target.key;
		if (key != NULL) key = fgw_strdup(key);
	}

	return channel_join(curr, chan, key);
}

static fgw_error_t act_part(fgw_arg_t *res, int argc, fgw_arg_t *argv)
{
	win_t *curr = idx2win(win_curr_idx);
	char *chan = NULL;
	gds_t msg = {0};
	int sid;
	fgw_error_t rv;

	ACTION_HELP((
		"part - irc command: part a channel",
		"Syntax:",
		"  part <#chan> [msg] - part #chan on win-associated server with part msg",
		"  part [msg]         - part current window's channel with part msg",
		NULL));

	ACT_MAY_CONVARG(1, FGW_STR, chan = argv[1].val.str);
	if (chan == NULL) { /* no channel, no message */
		if (curr->target.type != TARGET_CHANNEL) {
			ACT_ERROR("part", "Window is not a channel.");
			return FGW_ERR_ARG_CONV;
		}
		chan = curr->target.name;
		sid = curr->target.sid;
	}
	else if (!irc_is_prefix_chan(*chan)) { /* first arg does not look like a channel, use the windows' */
		if (curr->target.type != TARGET_CHANNEL) {
			ACT_ERROR("part", "Window is not a channel.");
			return FGW_ERR_ARG_CONV;
		}
		chan = curr->target.name;
		sid = curr->target.sid;
		ACT_MAY_GDS_REST(&msg, 1);
	}
	else { /* both chan and message are provided */
		ACT_MAY_GDS_REST(&msg, 2);
		sid = curr->target.sid;
	}

	rv = channel_part(sid, chan, msg.array);
	gds_uninit(&msg);
	return rv;
}

static fgw_error_t act_server(fgw_arg_t *res, int argc, fgw_arg_t *argv)
{
	char *name, *end, *sep, *hostname, tmp[128];
	int sid = -1, port = irc_default_port;
	win_t *curr = idx2win(win_curr_idx);
	target_type_t tt;
	gds_t str = {0};

	ACTION_HELP((
		"server - connect to an irc server",
		"Syntax:",
		"  server <id>/hostname[:port] - connect server, explicit id",
		"  server hostname[:port]      - connect server using current win's server id",
		"",
		"If server id is already in use, first disconnect from the current server",
		"and replace the connection with one to the newly named server for the",
		"same id.",
		NULL));

	no_change_window0();

	ACT_CONVARG(1, FGW_STR, name = argv[1].val.str);

	sep = strchr(name, '/');
	if (isdigit(name[0]) && (sep != NULL)) {
		sid = strtol(name, &end, 10);
		if (end != sep) {
			ACT_ERROR("server", "invalid server ID (must be a number)");
			return FGW_ERR_ARG_CONV;
		}
		if (sid < 0) {
			ACT_ERROR("server", "invalid server ID (out of range)");
			return FGW_ERR_ARG_CONV;
		}
		hostname = sep+1;
		tt = TARGET_SERVER_STATUS;
	}
	else {
		if (target_type_is_irc(curr->target.type)) {
			sid = curr->target.sid;
			tt = curr->target.type;
		}
		hostname = name;
	}

	if (sid < 0) {
		ACT_ERROR("server", "invalid server ID (can't determine existing sid)");
		return FGW_ERR_ARG_CONV;
	}

	hostname = fgw_strdup(hostname);
	sep = strchr(hostname, ':');
	if (sep != NULL) {
		*sep = '\0';
		sep++;
		port = strtol(sep, &end, 10);
		if (*end != '\0') {
			free(hostname);
			ACT_ERROR("server", "invalid server port (must be a number)");
			return FGW_ERR_ARG_CONV;
		}
		if ((port < 1) || (port> 65535)) {
			free(hostname);
			ACT_ERROR("server", "invalid server port (out of range)");
			return FGW_ERR_ARG_CONV;
		}
	}

	sid = irc_connect_new(curr, hostname, port, sid);
	if (sid < 0) {
		ACT_ERROR("server", "failed to connect");
		return FGW_ERR_ARG_CONV;
	}

	win_set_target(curr, tt, sid, "-", NULL);

	sprintf(tmp, "{%d/", sid); gds_append_str(&str, tmp);
	gds_append_str(&str, hostname);
	sprintf(tmp, ":%d}", port); gds_append_str(&str, tmp);

	free(curr->name);
	curr->name = str.array;

	curr->dirty = 1;
	free(hostname);
	return 0;
}

static fgw_error_t act_get_server(fgw_arg_t *res, int argc, fgw_arg_t *argv)
{
	const char *field, *rss = NULL;
	int sid = -1;
	irc_server_t *server;

	ACTION_HELP((
		"get_server - read fields of a server slot",
		"Syntax:",
		"  get_server nick [id]       - scripting: return own nick on a server (str)",
		"  get_server state [id]      - scripting: return server conn. state (str)",
		"  get_server host [id]       - scripting: return server hostname (str)",
		"  get_server port [id]       - scripting: return server port (int)",
		"  get_server status_win [id] - scripting: return status window id (int)",
		"",
		"If server id is not specified, use the current window's server.",
		NULL));

	ACT_CONVARG(1, FGW_STR, field = argv[1].val.str);
	ACT_MAY_CONVARG(2, FGW_INT, sid = argv[2].val.nat_int);

	if (sid < 0) {
		win_t *curr = idx2win(win_curr_idx);
		if (curr != NULL)
			sid = curr->target.sid;
	}

	if ((sid < 0) || (sid >= irc_servers.used)) {
		ACT_ERROR("get_server", "No such server.");
		return -1;
	}
	server = ((irc_server_t *)irc_servers.array[sid]);

	if (strcmp(field, "nick") == 0)            rss = server->nick;
	else if (strcmp(field, "host") == 0)       rss = server->hostname;
	else if (strcmp(field, "port") == 0) {
		res->type = FGW_INT;
		res->val.nat_int = server->port;
		return 0;
	}
	else if (strcmp(field, "status_win") == 0) {
		res->type = FGW_INT;
		res->val.nat_int = server->status_win->idx;
	}
	else if (strcmp(field, "state") == 0) {
		switch(server->state) {
			case IRCS_OFFLINE:         rss = "offline"; break;
			case IRCS_CONNECTING:      rss = "connecting"; break;
			case IRCS_CONNECTED:       rss = "connected"; break;
		}
	}
	else {
		ACT_ERROR("get_server", "Invalid field");
		return -1;
	}

	if (rss != NULL) {
		res->type = FGW_STR;
		res->val.cstr = rss;
	}

	return 0;
}


static fgw_error_t act_raw_(fgw_arg_t *res, int argc, fgw_arg_t *argv, int from, const char *colon_cmd, int sid)
{
	gds_t msg = {0};
	sockev_t s;

	if (sid < 0) {
		win_t *curr = idx2win(win_curr_idx);
		if (!target_type_is_irc(curr->target.type)) {
			ACT_ERROR("raw", "window is not associated with a server");
			return -1;
		}
		sid = curr->target.sid;
	}

	s = irc_server_sock(sid);
	if (s == SOCKEV_INVALID) {
		ACT_ERROR("raw", "server is not connected");
		return -1;
	}

	if (colon_cmd != NULL)
		sockev_puts(s, colon_cmd);

	if (colon_cmd != NULL)
		gds_append_str(&msg, " :");

	if (from == 0) {
		gds_append_str(&msg, argv[0].val.argv0.func->name);
		gds_append(&msg, ' ');
		if (argc > 1)
			ACT_GDS_REST(&msg, 1);
	}
	else
		ACT_GDS_REST(&msg, from);

	gds_append_str(&msg, "\r\n");
	sockev_puts(s, msg.array);
	gds_uninit(&msg);
	return 0;
}

static fgw_error_t act_raw(fgw_arg_t *res, int argc, fgw_arg_t *argv)
{
	ACTION_HELP((
		"raw - send raw irc protocol message to the server",
		"Syntax:",
		"  raw <irc_proto_line> ",
		NULL));
	return act_raw_(res, argc, argv, 1, NULL, -1);
}

static fgw_error_t act_rawto(fgw_arg_t *res, int argc, fgw_arg_t *argv)
{
	int sid;
	ACTION_HELP((
		"raw - send raw irc protocol message to the server",
		"Syntax:",
		"  raw <irc_proto_line> ",
		NULL));
	ACT_MAY_CONVARG(1, FGW_INT, sid = argv[1].val.nat_int);
	return act_raw_(res, argc, argv, 2, NULL, sid);
}

static fgw_error_t act_verbatim(fgw_arg_t *res, int argc, fgw_arg_t *argv)
{
	if ((argc > 1) && ((argv[1].type & FGW_STR) == FGW_STR) && (argv[1].val.str[0] == ACTION_HELP_STR[0]) && (strcmp(argv[1].val.str, ACTION_HELP_STR) == 0)) {
		char tmp[600]; /* assumes no irc command is longer than 510 chars */
		sprintf(tmp, "%s - as specified in the IRC protocol", argv[0].val.argv0.func->name);
		action_help_print(tmp, 0);
		return 0;
	}
	return act_raw_(res, argc, argv, 0, NULL, -1);
}

static fgw_error_t act_away(fgw_arg_t *res, int argc, fgw_arg_t *argv)
{
	ACTION_HELP((
		"away - set away state",
		"Syntax:",
		"  away <msg>     - activate away state, set away message",
		"  away           - deactivate away state, remove away message",
		NULL));
	return act_raw_(res, argc, argv, 1, "away", -1);
}


static fgw_error_t act_query(fgw_arg_t *res, int argc, fgw_arg_t *argv)
{
	char tmp[560];
	win_t *curr = idx2win(win_curr_idx);
	char *nick = NULL;

	ACTION_HELP((
		"query - irc command: use channel for private chat to a nick",
		"Syntax:",
		"  query <nick>      - chat with nick on server id associated with current window",
		"  query <id>/<nick> - chat with nick on server id ",
		"  query             - stop the chat, release window",
		NULL));

	no_change_window0();

	ACT_MAY_CONVARG(1, FGW_STR, nick = argv[1].val.str);
	if ((nick == NULL) || (*nick == '\0')) {
		if (curr->target.type != TARGET_NICK) {
			ACT_ERROR("query", "window is not a query to a nick");
			return -1;
		}
		free(curr->name);
		sprintf(tmp, "(queried %s)", curr->target.name);
		curr->name = fgw_strdup(tmp);
		win_set_target(curr, TARGET_NONE, curr->target.sid, win_target_keep, win_target_keep);
	}
	else {
		win_set_target(curr, TARGET_NICK, curr->target.sid, nick, win_target_keep);
		free(curr->name);
		sprintf(tmp, "[%s]", curr->target.name);
		curr->name = fgw_strdup(tmp);
	}

	curr->dirty = 1;
	return 0;
}

static fgw_error_t act_names(fgw_arg_t *res, int argc, fgw_arg_t *argv)
{
	win_t *curr = idx2win(win_curr_idx);
	char *chan = NULL, *target = "";
	sockev_t sock;

	ACTION_HELP((
		"names - irc command: list nicks on a channel",
		"Syntax:",
		"  names [#chan] - list nicks on #chan or channel associated with the window",
		NULL));

	ACT_MAY_CONVARG(1, FGW_STR, chan = argv[1].val.str);
	ACT_MAY_CONVARG(2, FGW_STR, target = argv[2].val.str);
	if (chan == NULL) {
		if (curr->target.type != TARGET_CHANNEL) {
			ACT_ERROR(argv[0].val.argv0.func->name, "Window is not a channel.");
			return FGW_ERR_ARG_CONV;
		}
		chan = curr->target.name;
	}

	sock = irc_server_sock(curr->target.sid);
	if (sock == SOCKEV_INVALID) {
		ACT_ERROR(argv[0].val.argv0.func->name, "Server not connected\n");
		return FGW_ERR_ARG_CONV;
	}
	sockev_printf(sock, "names %s %s\r\n", chan, target);
	return 0;
}

static fgw_error_t chan_rest(const char *cmd, const char *colon, fgw_arg_t *res, int argc, fgw_arg_t *argv)
{
	win_t *curr = idx2win(win_curr_idx);
	char *chan = NULL;
	gds_t rest = {0};
	sockev_t sock;

	ACT_MAY_CONVARG(1, FGW_STR, chan = argv[1].val.str);
	if ((chan == NULL) || (!irc_is_prefix_chan(*chan))) {
		if (curr->target.type != TARGET_CHANNEL) {
			ACT_ERROR(cmd, "Window is not a channel.");
			return FGW_ERR_ARG_CONV;
		}
		chan = curr->target.name;
		ACT_MAY_GDS_REST(&rest, 1);
	}
	else
		ACT_MAY_GDS_REST(&rest, 2);

	sock = irc_server_sock(curr->target.sid);
	if (sock == SOCKEV_INVALID) {
		ACT_ERROR(cmd, "Server not connected\n");
		return FGW_ERR_ARG_CONV;
	}
	if (rest.used == 0)
		sockev_printf(sock, "%s %s\r\n", cmd, chan);
	else
		sockev_printf(sock, "%s %s %s%s\r\n", cmd, chan, colon, rest.array);

	gds_uninit(&rest);
	return 0;
}

static fgw_error_t act_mode(fgw_arg_t *res, int argc, fgw_arg_t *argv)
{
	ACTION_HELP((
		"mode - irc command: query or change channel modes"
		"Syntax:",
		"  mode [#chan]       - query chan mode of #chan or current win's channel",
		"  mode [#chan] <mds> - set chan mode of #chan or current win's channel to mds",
		NULL));

	return chan_rest("mode", "", res, argc, argv);
}

static fgw_error_t act_topic(fgw_arg_t *res, int argc, fgw_arg_t *argv)
{
	ACTION_HELP((
		"topic - irc command: query or change channel topic"
		"Syntax:",
		"  topic [#chan]       - query topic of #chan or current win's channel",
		"  topic [#chan] <str> - set topic of #chan or current win's channel to str",
		NULL));
	return chan_rest("topic", ":", res, argc, argv);
}

static fgw_error_t who_whois(const char *cmd, fgw_arg_t *res, int argc, fgw_arg_t *argv)
{
	win_t *curr = idx2win(win_curr_idx);
	char *nick;
	sockev_t sock;

	ACT_CONVARG(1, FGW_STR, nick = argv[1].val.str);

	sock = irc_server_sock(curr->target.sid);
	if (sock == SOCKEV_INVALID) {
		ACT_ERROR(cmd, "Server not connected\n");
		return FGW_ERR_ARG_CONV;
	}
	sockev_printf(sock, "%s %s\r\n", cmd, nick);
	return 0;
}

static fgw_error_t act_whois(fgw_arg_t *res, int argc, fgw_arg_t *argv)
{
	ACTION_HELP((
		"whois - irc command: query whois info on a nick",
		"Syntax:",
		"  whois <nick>",
		NULL));

	return who_whois("whois", res, argc, argv);
}

static fgw_error_t act_who(fgw_arg_t *res, int argc, fgw_arg_t *argv)
{
	ACTION_HELP((
		"who - irc command: query who info on a pattern",
		"Syntax:",
		"  who <pattern>",
		NULL));
	return who_whois("who", res, argc, argv);
}



static fgw_error_t act_generic_channel_nick(fgw_arg_t *res, int argc, fgw_arg_t *argv, const char *cmd, int swap)
{
	win_t *curr = idx2win(win_curr_idx);
	char *chan = NULL, *nick = NULL;
	gds_t rest = {0};
	sockev_t sock;

	ACT_CONVARG(1, FGW_STR, chan = argv[1].val.str);
	if (!irc_is_prefix_chan(*chan)) {
		if (curr->target.type != TARGET_CHANNEL) {
			ACT_ERROR(cmd, "Window is not a channel.");
			return FGW_ERR_ARG_CONV;
		}
		nick = chan;
		chan = curr->target.name;
		ACT_MAY_GDS_REST(&rest, 2);
	}
	else {
		ACT_CONVARG(2, FGW_STR, nick = argv[2].val.str);
		ACT_MAY_GDS_REST(&rest, 3);
	}

	sock = irc_server_sock(curr->target.sid);
	if (sock == SOCKEV_INVALID) {
		ACT_ERROR(cmd, "Server not connected\n");
		return FGW_ERR_ARG_CONV;
	}

	if (swap) {
		char *tmp = chan;
		chan = nick;
		nick = tmp;
	}

	if (rest.used == 0)
		sockev_printf(sock, "%s %s %s\r\n", cmd, chan, nick);
	else
		sockev_printf(sock, "%s %s %s :%s\r\n", cmd, chan, nick, rest.array);

	gds_uninit(&rest);
	return 0;
}

static fgw_error_t act_kick(fgw_arg_t *res, int argc, fgw_arg_t *argv)
{
	ACTION_HELP((
		"kick - irc command: kick a nick from a channel",
		"Syntax:",
		"  kick [#chan] <nick> [message]",
		NULL));

	return act_generic_channel_nick(res, argc, argv, "kick", 0);
}

static fgw_error_t act_invite(fgw_arg_t *res, int argc, fgw_arg_t *argv)
{
	ACTION_HELP((
		"invite - irc command: invite a nick to a channel",
		"Syntax:",
		"  invite <nick> <#chan>",
		NULL));

	return act_generic_channel_nick(res, argc, argv, "invite", 1);
}

static fgw_error_t act_nick(fgw_arg_t *res, int argc, fgw_arg_t *argv)
{
	win_t *curr = idx2win(win_curr_idx);
	const char *nick, *cmd = "nick";
	sockev_t sock;

	ACT_CONVARG(1, FGW_STR, nick = argv[1].val.str);
	sock = irc_server_sock(curr->target.sid);
	if (sock == SOCKEV_INVALID) {
		ACT_ERROR(cmd, "Server not connected\n");
		return FGW_ERR_ARG_CONV;
	}
	sockev_printf(sock, "nick %s\r\n", nick);
	return 0;
}

static fgw_error_t act_msg_notice(fgw_arg_t *res, int argc, fgw_arg_t *argv)
{
	win_t *curr = idx2win(win_curr_idx);
	const char *target;
	int start, is_msg = argv[0].val.argv0.func->name[0] == 'm';
	gds_t rest = {0};
	sockev_t sock;

	ACTION_HELP((
		"msg & notice - send a message or notice to a nick or chan",
		"Syntax:",
		"  msg <nick|#chan> <message>",
		"  notice <nick|#chan> <message>",
		NULL));

	sock = irc_server_sock(curr->target.sid);
	if (sock == SOCKEV_INVALID) {
		ACT_ERROR(argv[0].val.argv0.func->name, "Server not connected\n");
		return FGW_ERR_ARG_CONV;
	}

	ACT_CONVARG(1, FGW_STR, target = argv[1].val.str);

	gds_append_str(&rest, is_msg ? ">> *" : ">> -");
	gds_append_str(&rest, target);
	gds_append_str(&rest, is_msg ? "* " : "- ");
	start = rest.used;
	ACT_MAY_GDS_REST(&rest, 2);

	sockev_printf(sock, "%s %s :%s\r\n", (is_msg ? "privmsg" : "notice"), target, rest.array + start);
	win_append(curr, rest.array);
	gds_uninit(&rest);
	return 0;
}


void irc_act_init(void)
{
	ircobj = fgw_obj_reg(&fgw, "irc");

	fgw_func_reg(ircobj, "join", act_join);
	fgw_func_reg(ircobj, "part", act_part);
	fgw_func_reg(ircobj, "get_server", act_get_server);
	fgw_func_reg(ircobj, "server", act_server);
	fgw_func_reg(ircobj, "raw", act_raw);
	fgw_func_reg(ircobj, "rawto", act_rawto);
	fgw_func_reg(ircobj, "quote", act_raw);
	fgw_func_reg(ircobj, "query", act_query);
	fgw_func_reg(ircobj, "names", act_names);
	fgw_func_reg(ircobj, "mode", act_mode);
	fgw_func_reg(ircobj, "topic", act_topic);
	fgw_func_reg(ircobj, "whois", act_whois);
	fgw_func_reg(ircobj, "who", act_who);
	fgw_func_reg(ircobj, "kick", act_kick);
	fgw_func_reg(ircobj, "nick", act_nick);
	fgw_func_reg(ircobj, "away", act_away);
	fgw_func_reg(ircobj, "invite", act_invite);
	fgw_func_reg(ircobj, "msg", act_msg_notice);
	fgw_func_reg(ircobj, "notice", act_msg_notice);


	/* these can be passed on without modification */
	fgw_func_reg(ircobj, "links", act_verbatim);
	fgw_func_reg(ircobj, "whowas", act_verbatim);
	fgw_func_reg(ircobj, "list", act_verbatim);
}
