/*  fgwirc - fungw based minimal irc client
    Copyright (C) 2019,2021  Tibor 'Igor2' Palinkas

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along
    with this program; if not, write to the Free Software Foundation, Inc.,
    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

    Project page and contact: http://repo.hu/projects/fgwirc
*/

#include "action.h"
#include "window.h"
#include "irc.h"

#include "window_act.h"

static fgw_obj_t *winobj;

static fgw_error_t act_win(fgw_arg_t *res, int argc, fgw_arg_t *argv)
{
	win_t *curr = idx2win(win_curr_idx);
	char *cmd = NULL;
	int idx;

	if (argv[0].val.argv0.func->name[3] == 'd')
		ACTION_HELP(("window - alias to win", NULL));

	ACTION_HELP((
		"win - various window operations (UI)",
		"Syntax: ",
		"  win <number>        - switch to and show window number (absolute)",
		"  win [show|swap] <N> - same as above",
		"  win <+number>       - switch to and show window number (relative)",
		"  win <-number>       - switch to and show window number (relative)",
		"  win new             - create (open) a new window and switch to it",
		"  win new*<number>    - create (open) multiple new windows",
		"  win server <id>     - associate current window to integer server id",
		"  win server          - return current window's server id association",
		"  win release         - release (remove) association of current window",
		"  win channel <#chan> - show #chan (on server last assigned) in this win",
		"  win redraw          - redraw the screen (current window)",
		"  win log on <path>   - start logging current window to file at path (append)",
		"  win log off         - stop logging current window",
		"  win size            - echo current window size",
		"  win resize          - query terminal size and set matching window size",
		"  win resize <w> <h>  - set window size to w (width) and h (height), integers",
		"  win append <id> <t> - append text t at the end (bottom) of window id",
		"",
		"scripting:",
		"",
		"  win get num         - return number of windows",
		"  win get current     - return id of currently visible window",
		"  win get type <id>   - return window target type",
		"  win get target <id> - return window target name (nick or chan)",
		"  win get srvid <id>  - return window target server ID",
		"  win get dirty <id>  - return 0 or 1 if window has unseen activity",
		"",
		"  win set dirty <id> <val>  - set or clear window unseen activity (1 or 0)",
		NULL));


	ACT_MAY_CONVARG(1, FGW_STR, cmd = argv[1].val.str);

	if (cmd == NULL) {
		char tmp[1024], *end = tmp;

		end += sprintf(end, "Window %d: ", win_curr_idx);
		switch(curr->target.type) {
			case TARGET_NONE:            end += sprintf(end, "none"); break;
			case TARGET_CHANNEL:         end += sprintf(end, "channel %s (key=%s) %s", curr->target.name, curr->target.key == NULL ? "" : curr->target.key, curr->target.state == TGTST_NORMAL ? "joined" : "not joined"); break;
			case TARGET_NICK:            end += sprintf(end, "query %s", curr->target.name); break;
			case TARGET_SERVER_STATUS:   end += sprintf(end, "server status"); break;
			case TARGET_FGWIRC_STATUS:   end += sprintf(end, "fgwirc status"); break;
		}
		if (target_type_is_irc(curr->target.type))
			end += sprintf(end, " on network %d", curr->target.sid);
		win_append(curr, tmp);
		return 0;
	}

	if ((*cmd >= '0') && (*cmd <= '9')) {
		win_show_idx(atoi(cmd));
		return 0;
	}

	if (((cmd[0] == '+') || (cmd[0] == '-')) && ((cmd[1] >= '0') && (cmd[1] <= '9'))) {
		int nidx = (win_curr_idx + atoi(cmd));
		if (nidx < 0)
			nidx = (-((-(nidx)-1) % windows.used)-1) + windows.used;
		else
			nidx = nidx % windows.used;
		win_show_idx(nidx);
		return 0;
	}

	if ((strcmp(cmd, "show") == 0) || (strcmp(cmd, "swap") == 0)) {
		ACT_CONVARG(2, FGW_INT, idx = argv[2].val.nat_int);
		win_show_idx(idx);
		return 0;
	}

	if (strcmp(cmd, "new") == 0) {
		win_t *w = win_new("-");
		if (curr->target.sid >= 0)
			win_set_target_sid(w, curr->target.sid);
		win_show(w);
		return 0;
	}

	if (strncmp(cmd, "new*", 4) == 0) {
		win_t *w;
		int rep;

		for(rep = atoi(cmd+4); rep > 0; rep--) {
			w = win_new("-");
			if (curr->target.sid >= 0)
				win_set_target_sid(w, curr->target.sid);
		}
		win_show(w);
		return 0;
	}


	if (strcmp(cmd, "server") == 0) {
		if (argc > 2) {
			char *name;
			int sid;

			no_change_window0();

			ACT_CONVARG(2, FGW_STR, name = argv[2].val.str);
			sid = irc_lookup_server(name);
			if (sid < 0) {
				ACT_ERROR("win", "no such server");
				return 1;
			}
			win_set_target(curr, TARGET_NONE, sid, win_target_keep, win_target_keep);
			free(curr->name); curr->name = fgw_strdup("-");
			curr->dirty = 1;
		}
		else { /* query server id */
			res->type = FGW_INT;
			res->val.nat_int = curr->target.sid;
		}
		return 0;
	}

	if (strcmp(cmd, "release") == 0) {
		no_change_window0();
		win_set_target(curr, TARGET_NONE, curr->target.sid, win_target_keep, win_target_keep);
		free(curr->name); curr->name = fgw_strdup("-");
		curr->dirty = 1;
		return 0;
	}

	if (strcmp(cmd, "channel") == 0) {
		char *name, *key = NULL;
		ACT_CONVARG(2, FGW_STR, name = argv[2].val.str);
		ACT_MAY_CONVARG(3, FGW_STR, key = argv[3].val.str);
		win_set_target(curr, TARGET_CHANNEL, curr->target.sid, name, key);
		return 0;
	}

	if (strcmp(cmd, "append") == 0) {
		char *msg;
		int wid;
		win_t *w;

		ACT_CONVARG(2, FGW_INT, wid = argv[2].val.nat_int);
		ACT_CONVARG(3, FGW_STR, msg = argv[3].val.str);

		w = idx2win(wid);
		if (w == NULL)
			return -1;

		win_append(w, msg);
		return 0;
	}

	if (strcmp(cmd, "redraw") == 0) {
		curr->dirty = 1;
		return 0;
	}

	if (strcmp(cmd, "log") == 0) {
		const char *r, *op, *name;

		ACT_CONVARG(2, FGW_STR, op = argv[2].val.str);
		if (strcmp(op, "on") == 0) {
			ACT_CONVARG(3, FGW_STR, name = argv[3].val.str);
			r = win_log(curr, name, 1);
			if (r != NULL)
				win_append(curr, r);
			else
				win_append(curr, "Window log started.");
		}
		if (strcmp(op, "off") == 0) {
			r = win_log(curr, NULL, 0);
			if (r != NULL)
				win_append(curr, r);
			else
				win_append(curr, "Window log stopped.");
		}
		return 0;
	}

	if (strcmp(cmd, "size") == 0) {
		char tmp[256];
		sprintf(tmp, "Window size: %d cols %d lines", term->sx, term->sy);
		win_append(curr, tmp);
		return 0;
	}

	if (strcmp(cmd, "resize") == 0) {
		int sx = 0, sy = 0;
		ACT_MAY_CONVARG(2, FGW_INT, sx = argv[2].val.nat_int);
		ACT_MAY_CONVARG(3, FGW_INT, sy = argv[3].val.nat_int);
		if ((sx > 512) || (sy > 512)) {
			ACT_ERROR("win", "window too large");
			return 1;
		}
		win_resize(sx, sy);
		return 0;
	}

	if (strcmp(cmd, "get") == 0) {
		ACT_MAY_CONVARG(2, FGW_STR, cmd = argv[2].val.str);

		if (strcmp(cmd, "num") == 0) {
			res->type = FGW_INT; res->val.nat_int = windows.used;
			return 0;
		}
		if (strcmp(cmd, "current") == 0) {
			res->type = FGW_INT; res->val.nat_int = win_curr_idx;
			return 0;
		}

		/* for quieries requiring a window idx */
		ACT_CONVARG(3, FGW_INT, idx = argv[3].val.nat_int);
		if ((idx < 0) || (idx >= windows.used)) return 1;

		if (strcmp(cmd, "srvid") == 0) {
			res->type = FGW_INT; res->val.nat_int = ((win_t *)windows.array[idx])->target.sid;
			return 0;
		}

		if (strcmp(cmd, "dirty") == 0) {
			res->type = FGW_INT; res->val.nat_int = ((win_t *)windows.array[idx])->dirty;
			return 0;
		}

		if ((strcmp(cmd, "type") == 0) || (strcmp(cmd, "target") == 0)) {
			res->type = FGW_STR;
			switch(((win_t *)windows.array[idx])->target.type) {
				case TARGET_NONE:            res->val.cstr = (cmd[1] == 'y') ? "none" : ""; break;
				case TARGET_CHANNEL:         res->val.cstr = (cmd[1] == 'y') ? "channel" : ((win_t *)windows.array[idx])->target.name; break;
				case TARGET_NICK:            res->val.cstr = (cmd[1] == 'y') ? "query" : ((win_t *)windows.array[idx])->target.name; break;
				case TARGET_SERVER_STATUS:   res->val.cstr = (cmd[1] == 'y') ? "server-status" : ""; break;
				case TARGET_FGWIRC_STATUS:   res->val.cstr = (cmd[1] == 'y') ? "fgwirc-status" : ""; break;
				default: return -1;
			}
			return 0;
		}
	}

	if (strcmp(cmd, "set") == 0) {
		ACT_MAY_CONVARG(2, FGW_STR, cmd = argv[2].val.str);

		/* for sets requiring a window idx */
		ACT_CONVARG(3, FGW_INT, idx = argv[3].val.nat_int);
		if ((idx < 0) || (idx >= windows.used)) return 1;

		if (strcmp(cmd, "dirty") == 0) {
			ACT_CONVARG(4, FGW_INT, ((win_t *)windows.array[idx])->dirty = argv[4].val.nat_int);
			return 0;
		}
	}


	ACT_ERROR("win", "invalid cmd");
	return 1;
}

static fgw_error_t act_scroll(fgw_arg_t *res, int argc, fgw_arg_t *argv)
{
	win_t *curr = idx2win(win_curr_idx);
	int offs;

	ACTION_HELP((
		"scroll - scroll current window",
		"Syntax: scroll [offs]",
		"",
		"Scroll current window by positive (down) or negative (up) offset (integer).",
		"If offset is 0, scroll to the end (bottom).",
		NULL));

	ACT_CONVARG(1, FGW_INT, offs = argv[1].val.nat_int);

	if (offs < 0)
		win_scroll_up(curr, -offs);
	else if (offs > 0)
		win_scroll_down(curr, offs);
	else
		curr->scroll_start = NULL;
	curr->dirty = 1;
	return 0;
}

void window_act_init(void)
{
	winobj = fgw_obj_reg(&fgw, "window");

	fgw_func_reg(winobj, "win", act_win);
	fgw_func_reg(winobj, "window", act_win);
	fgw_func_reg(winobj, "scroll", act_scroll);
}
